@@ -1,7 +1,5 @@
 #include "pgroonga.h"
 
-#include "pgrn-compatible.h"
-
 #include "pgrn-global.h"
 #include "pgrn-groonga.h"
 #include "pgrn-options.h"
@@ -10,7 +8,8 @@
 
 #include <xxhash.h>
 
-typedef enum {
+typedef enum
+{
 	PGRN_SEQUENTIAL_SEARCH_TARGET_TEXT,
 	PGRN_SEQUENTIAL_SEARCH_TARGET_TEXTS,
 } PGrnSequentialSearchTargetType;
@@ -40,11 +39,11 @@ typedef struct PGrnSequentialSearchDatum
 	grn_obj *expression;
 	grn_obj *variable;
 	bool useIndex;
+	float4 fuzzyMaxDistanceRatio;
 	grn_expr_flags exprFlags;
 	bool used;
 } PGrnSequentialSearchDatum;
 
-static grn_ctx *ctx = &PGrnContext;
 static struct PGrnBuffers *buffers = &PGrnBuffers;
 static PGrnSequentialSearchTargetType currentTargetType =
 	PGRN_SEQUENTIAL_SEARCH_TARGET_TEXT;
@@ -58,11 +57,8 @@ static void
 PGrnSequentialSearchDatumInitialize(PGrnSequentialSearchDatum *datum)
 {
 	grn_column_flags flags = 0;
-	datum->table = grn_table_create(ctx,
-									NULL, 0,
-									NULL,
-									GRN_OBJ_TABLE_NO_KEY,
-									NULL, NULL);
+	datum->table =
+		grn_table_create(ctx, NULL, 0, NULL, GRN_OBJ_TABLE_NO_KEY, NULL, NULL);
 	datum->targetType = currentTargetType;
 	if (datum->targetType == PGRN_SEQUENTIAL_SEARCH_TARGET_TEXT)
 	{
@@ -74,20 +70,19 @@ PGrnSequentialSearchDatumInitialize(PGrnSequentialSearchDatum *datum)
 	}
 	datum->targetColumn = grn_column_create(ctx,
 											datum->table,
-											"target", strlen("target"),
+											"target",
+											strlen("target"),
 											NULL,
 											flags,
 											grn_ctx_at(ctx, GRN_DB_TEXT));
-	datum->recordID = grn_table_add(ctx,
-									datum->table,
-									NULL, 0,
-									NULL);
+	datum->recordID = grn_table_add(ctx, datum->table, NULL, 0, NULL);
 	datum->indexOID = InvalidOid;
 	datum->lexicon = NULL;
 	datum->indexColumn = NULL;
 	datum->matched =
 		grn_table_create(ctx,
-						 NULL, 0,
+						 NULL,
+						 0,
 						 NULL,
 						 GRN_OBJ_TABLE_HASH_KEY | GRN_OBJ_WITH_SUBREC,
 						 datum->table,
@@ -97,6 +92,7 @@ PGrnSequentialSearchDatumInitialize(PGrnSequentialSearchDatum *datum)
 	datum->expression = NULL;
 	datum->variable = NULL;
 	datum->useIndex = false;
+	datum->fuzzyMaxDistanceRatio = 0.0;
 	datum->exprFlags = PGRN_EXPR_QUERY_PARSE_FLAGS;
 	datum->used = true;
 }
@@ -130,13 +126,15 @@ PGrnInitializeSequentialSearch(void)
 void
 PGrnFinalizeSequentialSearch(void)
 {
-	GRN_HASH_EACH_BEGIN(ctx, data, cursor, id) {
+	GRN_HASH_EACH_BEGIN(ctx, data, cursor, id)
+	{
 		void *value;
 		PGrnSequentialSearchDatum *datum;
 		grn_hash_cursor_get_value(ctx, cursor, &value);
 		datum = value;
 		PGrnSequentialSearchDatumFinalize(datum);
-	} GRN_HASH_EACH_END(ctx, cursor);
+	}
+	GRN_HASH_EACH_END(ctx, cursor);
 	grn_hash_close(ctx, data);
 }
 
@@ -160,7 +158,8 @@ PGrnReleaseSequentialSearch(ResourceReleasePhase phase,
 	}
 
 	GRN_LOG(ctx, GRN_LOG_DEBUG, "%s[start] %u", tag, grn_hash_size(ctx, data));
-	GRN_HASH_EACH_BEGIN(ctx, data, cursor, id) {
+	GRN_HASH_EACH_BEGIN(ctx, data, cursor, id)
+	{
 		void *value;
 		PGrnSequentialSearchDatum *datum;
 		grn_hash_cursor_get_value(ctx, cursor, &value);
@@ -178,13 +177,13 @@ PGrnReleaseSequentialSearch(ResourceReleasePhase phase,
 			PGrnSequentialSearchDatumFinalize(datum);
 			grn_hash_cursor_delete(ctx, cursor, NULL);
 		}
-	} GRN_HASH_EACH_END(ctx, cursor);
+	}
+	GRN_HASH_EACH_END(ctx, cursor);
 	GRN_LOG(ctx, GRN_LOG_DEBUG, "%s[end] %u", tag, grn_hash_size(ctx, data));
 }
 
 void
-PGrnSequentialSearchSetTargetText(const char *target,
-								  unsigned int targetSize)
+PGrnSequentialSearchSetTargetText(const char *target, unsigned int targetSize)
 {
 	grn_obj *text = &(buffers->text);
 	GRN_TEXT_SET(ctx, text, target, targetSize);
@@ -192,8 +191,7 @@ PGrnSequentialSearchSetTargetText(const char *target,
 }
 
 void
-PGrnSequentialSearchSetTargetTexts(ArrayType *targets,
-								   PGrnCondition *condition)
+PGrnSequentialSearchSetTargetTexts(ArrayType *targets, PGrnCondition *condition)
 {
 	grn_obj *texts = &(buffers->texts);
 	ArrayIterator iterator;
@@ -210,7 +208,7 @@ PGrnSequentialSearchSetTargetTexts(ArrayType *targets,
 	{
 		const char *target = NULL;
 		unsigned int targetSize = 0;
-		unsigned int weight  = 0;
+		unsigned int weight = 0;
 		grn_id domain = GRN_DB_TEXT;
 
 		if (nTargets > i && !GRN_BOOL_VALUE_AT(condition->isTargets, i))
@@ -219,19 +217,12 @@ PGrnSequentialSearchSetTargetTexts(ArrayType *targets,
 		if (isNULL)
 			continue;
 
-		PGrnPGDatumExtractString(datum,
-								 ARR_ELEMTYPE(targets),
-								 &target,
-								 &targetSize);
+		PGrnPGDatumExtractString(
+			datum, ARR_ELEMTYPE(targets), &target, &targetSize);
 		if (!target)
 			continue;
 
-		grn_vector_add_element(ctx,
-							   texts,
-							   target,
-							   targetSize,
-							   weight,
-							   domain);
+		grn_vector_add_element(ctx, texts, target, targetSize, weight, domain);
 	}
 	array_free_iterator(iterator);
 
@@ -301,8 +292,7 @@ PGrnSequentialSearchPrepareIndex(PGrnCondition *condition,
 	key.useIndex = (PGrnIsTemporaryIndexSearchAvailable &&
 					(OidIsValid(indexOID) || targetIsVector));
 	key.type = type;
-	if (currentDatum &&
-		currentDatum->indexOID == key.indexOID &&
+	if (currentDatum && currentDatum->indexOID == key.indexOID &&
 		currentDatum->attributeIndex == key.attributeIndex &&
 		currentDatum->targetType == key.targetType &&
 		currentDatum->useIndex == key.useIndex &&
@@ -332,6 +322,7 @@ PGrnSequentialSearchPrepareIndex(PGrnCondition *condition,
 	currentDatum->attributeIndex = key.attributeIndex;
 	currentDatum->useIndex = key.useIndex;
 	currentDatum->type = key.type;
+	currentDatum->fuzzyMaxDistanceRatio = condition->fuzzyMaxDistanceRatio;
 	if (!currentDatum->useIndex)
 	{
 		return true;
@@ -347,7 +338,8 @@ PGrnSequentialSearchPrepareIndex(PGrnCondition *condition,
 
 		index = PGrnPGResolveIndexID(currentDatum->indexOID);
 		isPGroongaIndex = PGrnIndexIsPGroonga(index);
-		if (!isPGroongaIndex) {
+		if (!isPGroongaIndex)
+		{
 			PGrnSequentialSearchDatumFinalize(currentDatum);
 			grn_hash_delete(ctx, data, &key, sizeof(key), NULL);
 			currentDatum = NULL;
@@ -355,20 +347,20 @@ PGrnSequentialSearchPrepareIndex(PGrnCondition *condition,
 			PGrnCheckRC(GRN_INVALID_ARGUMENT,
 						"%s[invalid] not PGroonga index: <%.*s%s%.*s%s%.*s>",
 						tag,
-						(int) schemaNameSize, schemaName,
+						(int) schemaNameSize,
+						schemaName,
 						schemaNameSize > 0 ? "." : "",
-						(int) indexNameSize, indexName,
+						(int) indexNameSize,
+						indexName,
 						columnNameSize > 0 ? "." : "",
-						(int) columnNameSize, columnName);
+						(int) columnNameSize,
+						columnName);
 		}
 
 		PG_TRY();
 		{
-			currentDatum->lexicon =
-				PGrnCreateSimilarTemporaryLexicon(index,
-												  columnName,
-												  columnNameSize,
-												  tag);
+			currentDatum->lexicon = PGrnCreateSimilarTemporaryLexicon(
+				index, columnName, columnNameSize, tag);
 			currentDatum->exprFlags |= PGrnOptionsGetExprParseFlags(index);
 		}
 		PG_CATCH();
@@ -382,22 +374,19 @@ PGrnSequentialSearchPrepareIndex(PGrnCondition *condition,
 		PG_END_TRY();
 		RelationClose(index);
 
-		if (grn_obj_get_info(ctx,
-							 currentDatum->lexicon,
-							 GRN_INFO_DEFAULT_TOKENIZER,
-							 NULL))
+		if (grn_obj_get_info(
+				ctx, currentDatum->lexicon, GRN_INFO_DEFAULT_TOKENIZER, NULL))
 		{
 			indexFlags |= GRN_OBJ_WITH_POSITION;
 		}
 
 		PG_TRY();
 		{
-			currentDatum->indexColumn =
-				PGrnCreateColumn(InvalidRelation,
-								 currentDatum->lexicon,
-								 "index",
-								 indexFlags,
-								 currentDatum->table);
+			currentDatum->indexColumn = PGrnCreateColumn(InvalidRelation,
+														 currentDatum->lexicon,
+														 "index",
+														 indexFlags,
+														 currentDatum->table);
 			PGrnIndexColumnSetSource(InvalidRelation,
 									 currentDatum->indexColumn,
 									 currentDatum->targetColumn);
@@ -438,12 +427,11 @@ PGrnSequentialSearchPrepareIndex(PGrnCondition *condition,
 								tokenizer,
 								normalizers,
 								NULL);
-			currentDatum->indexColumn =
-				PGrnCreateColumn(InvalidRelation,
-								 currentDatum->lexicon,
-								 "index",
-								 indexFlags,
-								 currentDatum->table);
+			currentDatum->indexColumn = PGrnCreateColumn(InvalidRelation,
+														 currentDatum->lexicon,
+														 "index",
+														 indexFlags,
+														 currentDatum->table);
 			PGrnIndexColumnSetSource(InvalidRelation,
 									 currentDatum->indexColumn,
 									 currentDatum->targetColumn);
@@ -488,9 +476,8 @@ PGrnSequentialSearchPrepareExpression(PGrnCondition *condition,
 							  currentDatum->variable);
 	if (!currentDatum->expression)
 	{
-		PGrnCheckRC(GRN_NO_MEMORY_AVAILABLE,
-					"%s failed to create expression",
-					tag);
+		PGrnCheckRC(
+			GRN_NO_MEMORY_AVAILABLE, "%s failed to create expression", tag);
 	}
 
 	currentDatum->expressionHash = expressionHash;
@@ -503,30 +490,25 @@ PGrnSequentialSearchSetMatchTerm(PGrnCondition *condition)
 {
 	const char *tag = "[sequential-search][match-term]";
 
-	if (PGrnSequentialSearchPrepareExpression(condition,
-											  PGRN_SEQUENTIAL_SEARCH_MATCH_TERM))
+	if (PGrnSequentialSearchPrepareExpression(
+			condition, PGRN_SEQUENTIAL_SEARCH_MATCH_TERM))
 	{
 		return;
 	}
 
-	grn_expr_append_obj(ctx,
-						currentDatum->expression,
-						currentDatum->targetColumn,
-						GRN_OP_GET_VALUE,
-						1);
-	PGrnCheck("%s append match target column", tag);
-	grn_expr_append_const_str(ctx,
-							  currentDatum->expression,
+	PGrnExprAppendObject(currentDatum->expression,
+						 currentDatum->targetColumn,
+						 GRN_OP_GET_VALUE,
+						 1,
+						 tag,
+						 NULL);
+	PGrnExprAppendConstString(currentDatum->expression,
 							  VARDATA_ANY(condition->query),
 							  VARSIZE_ANY_EXHDR(condition->query),
 							  GRN_OP_PUSH,
-							  1);
-	PGrnCheck("%s append term to be matched", tag);
-	grn_expr_append_op(ctx,
-					   currentDatum->expression,
-					   GRN_OP_MATCH,
-					   2);
-	PGrnCheck("%s append match operator", tag);
+							  1,
+							  tag);
+	PGrnExprAppendOp(currentDatum->expression, GRN_OP_MATCH, 2, tag, NULL);
 }
 
 void
@@ -534,30 +516,25 @@ PGrnSequentialSearchSetEqualText(PGrnCondition *condition)
 {
 	const char *tag = "[sequential-search][equal-text]";
 
-	if (PGrnSequentialSearchPrepareExpression(condition,
-											  PGRN_SEQUENTIAL_SEARCH_EQUAL_TEXT))
+	if (PGrnSequentialSearchPrepareExpression(
+			condition, PGRN_SEQUENTIAL_SEARCH_EQUAL_TEXT))
 	{
 		return;
 	}
 
-	grn_expr_append_obj(ctx,
-						currentDatum->expression,
-						currentDatum->targetColumn,
-						GRN_OP_GET_VALUE,
-						1);
-	PGrnCheck("%s append match target column", tag);
-	grn_expr_append_const_str(ctx,
-							  currentDatum->expression,
+	PGrnExprAppendObject(currentDatum->expression,
+						 currentDatum->targetColumn,
+						 GRN_OP_GET_VALUE,
+						 1,
+						 tag,
+						 NULL);
+	PGrnExprAppendConstString(currentDatum->expression,
 							  VARDATA_ANY(condition->query),
 							  VARSIZE_ANY_EXHDR(condition->query),
 							  GRN_OP_PUSH,
-							  1);
-	PGrnCheck("%s append equal text", tag);
-	grn_expr_append_op(ctx,
-					   currentDatum->expression,
-					   GRN_OP_EQUAL,
-					   2);
-	PGrnCheck("%s append equal operator", tag);
+							  1,
+							  tag);
+	PGrnExprAppendOp(currentDatum->expression, GRN_OP_EQUAL, 2, tag, NULL);
 }
 
 void
@@ -571,24 +548,19 @@ PGrnSequentialSearchSetPrefix(PGrnCondition *condition)
 		return;
 	}
 
-	grn_expr_append_obj(ctx,
-						currentDatum->expression,
-						currentDatum->targetColumn,
-						GRN_OP_GET_VALUE,
-						1);
-	PGrnCheck("%s append match target column", tag);
-	grn_expr_append_const_str(ctx,
-							  currentDatum->expression,
+	PGrnExprAppendObject(currentDatum->expression,
+						 currentDatum->targetColumn,
+						 GRN_OP_GET_VALUE,
+						 1,
+						 tag,
+						 NULL);
+	PGrnExprAppendConstString(currentDatum->expression,
 							  VARDATA_ANY(condition->query),
 							  VARSIZE_ANY_EXHDR(condition->query),
 							  GRN_OP_PUSH,
-							  1);
-	PGrnCheck("%s append prefix", tag);
-	grn_expr_append_op(ctx,
-					   currentDatum->expression,
-					   GRN_OP_PREFIX,
-					   2);
-	PGrnCheck("%s append prefix operator", tag);
+							  1,
+							  tag);
+	PGrnExprAppendOp(currentDatum->expression, GRN_OP_PREFIX, 2, tag, NULL);
 }
 
 void
@@ -614,9 +586,8 @@ PGrnSequentialSearchSetQuery(PGrnCondition *condition,
 				   currentDatum->exprFlags);
 	if (ctx->rc != GRN_SUCCESS)
 		currentDatum->expressionHash = 0;
-	PGrnCheck("%s failed to parse expression: <%.*s>",
-			  tag,
-			  (int) querySize, query);
+	PGrnCheck(
+		"%s failed to parse expression: <%.*s>", tag, (int) querySize, query);
 }
 
 void
@@ -638,13 +609,13 @@ PGrnSequentialSearchSetScript(PGrnCondition *condition)
 				   script,
 				   scriptSize,
 				   currentDatum->targetColumn,
-				   GRN_OP_MATCH, GRN_OP_AND,
+				   GRN_OP_MATCH,
+				   GRN_OP_AND,
 				   flags);
 	if (ctx->rc != GRN_SUCCESS)
 		currentDatum->expressionHash = 0;
-	PGrnCheck("%s failed to parse expression: <%.*s>",
-			  tag,
-			  (int) scriptSize, script);
+	PGrnCheck(
+		"%s failed to parse expression: <%.*s>", tag, (int) scriptSize, script);
 }
 
 bool
@@ -669,11 +640,12 @@ PGrnSequentialSearchExecute(void)
 
 	if (currentDatum->useIndex)
 	{
-		grn_table_select(ctx,
-						 currentDatum->table,
-						 currentDatum->expression,
-						 currentDatum->matched,
-						 GRN_OP_OR);
+		grn_table_selector *table_selector = grn_table_selector_open(
+			ctx, currentDatum->table, currentDatum->expression, GRN_OP_OR);
+		grn_table_selector_set_fuzzy_max_distance_ratio(
+			ctx, table_selector, currentDatum->fuzzyMaxDistanceRatio);
+		grn_table_selector_select(ctx, table_selector, currentDatum->matched);
+		grn_table_selector_close(ctx, table_selector);
 
 		if (grn_table_size(ctx, currentDatum->matched) == 1)
 		{
